-- test to see just how bad older profiles get mangled (or not). They likely get completely trashed. Consider using new profile, might be good for people reverting
-- finish localization
-- create the simple bar display, and potentially allow it to merge somehow

local AddonName, Hermes = ...

local LIB_AceAddon = LibStub("AceAddon-3.0") or error("Required library AceAddon-3.0 not found")
_G[AddonName] = LIB_AceAddon:NewAddon(Hermes, AddonName, "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceComm-3.0", "AceSerializer-3.0")
local L = LibStub('AceLocale-3.0'):GetLocale('Hermes')
local LIB_AceConfigRegistry = LibStub("AceConfigRegistry-3.0") or error("Required library AceConfigRegistry-3.0 not found")

--default all modules to disabled
Hermes:SetDefaultModuleState(false)

local dbp

Hermes.Player = {}
Hermes.Senders = {}
Hermes.Abilities = {}
Hermes.AbilityInstances = {}

--------------------------------------------------------------------
-- Static Data Structures
--------------------------------------------------------------------
--local independent class lookup, used mostly to keep message size down when sending messages.
Hermes.CLASS_ENUM = {
	[1]		= "ANY",
	[2]		= "DEATHKNIGHT",
	[3]		= "DRUID",
	[4]		= "HUNTER",
	[5]		= "MAGE",
	[6]		= "PALADIN",
	[7]		= "PRIEST",
	[8]		= "ROGUE",
	[9]		= "SHAMAN",
	[10]	= "WARLOCK",
	[11]	= "WARRIOR",
}

local MESSAGE_ENUM = {
	[1]		= "INITIALIZE_SENDER",
	[2]		= "INITIALIZE_RECEIVER",
	[3]		= "REQUEST_SPELLS",
	[4]		= "UPDATE_SPELLS",
}

HERMES_SEND_COMM = "HermesS1"
HERMES_RECEIVE_COMM = "HermesR1"

Hermes.Sender = nil
Hermes.Receiving = false

local ItemNameTimer = nil
local ITEM_NAME_THROTTLE = 20									--how often to scan server for itme name info, three times a minute currently

local CooldownScanTimer = nil
local COOLDOWN_SCAN_FREQUENCY_INITIAL = 10						--Used the throttle oneself when first starting up sending.
																--This is to allow time for all the receivers in the raid to request their spells.
																--This results in fewer SPELL_UPDATES being sent when first starting up
local COOLDOWN_SCAN_FREQUENCY = 1
local COOLDOWN_DELTA_THRESHOLD = COOLDOWN_SCAN_FREQUENCY * 5	--used to avoid sending messages for cooldowns that have expired normally
																--and used to determine whether cooldowns have decreased unexpectedly
																--(such as from tier set bonusus or anything else that might reduce a cooldown)
Hermes.ITEM_NOT_IN_IVENTORY_OR_EQUIPPED = -50000				--arbritrary value to indicate when a sender doesn't have an item

local localizedClassNames = {}
FillLocalizedClassList(localizedClassNames)

--specila handling for soulstone
local SPELLID_SOULSTONERESURRECTION = 20707
local ITEMID_SOULSTONE = -5232
local STARTTIME_SOULSTONERESURRECTION = nil

--------------------------------------------------------------------
-- WoW Events
--------------------------------------------------------------------
local PLAYER_ENTERED_WORLD = nil
function Hermes:PLAYER_ENTERING_WORLD()				--used only for one time player initialization that has to happen after player is logged in.
	self:UnregisterEvent("PLAYER_ENTERING_WORLD")
	
	--initialize static properties for player
	Hermes:InitializePlayer()
	
	PLAYER_ENTERED_WORLD = true
	
	--New changes required to support patch 4.1 and RegisterAddonMessagePrefix
	--this code should be updated sometime after 4.1 ships
	if RegisterAddonMessagePrefix then
		if RegisterAddonMessagePrefix(HERMES_SEND_COMM) == false then
			--register failed for some reason
			local message = L["|cFFFF0000Hermes Warning|r"].." ".."Failed to register addon prefix: "..HERMES_SEND_COMM
			print(message)
		end
		
		if RegisterAddonMessagePrefix(HERMES_RECEIVE_COMM) == false then
			--register failed for some reason
			local message = L["|cFFFF0000Hermes Warning|r"].." ".."Failed to register addon prefix: "..HERMES_RECEIVE_COMM
			print(message)
		end
	end
	
	if(dbp.configMode == true) then
		--show the warning message
		self:ShowHermesTestModeMessage()
		self:StartTestMode()
	else
		self:UpdateCommunicationsStatus()
	end
end

function Hermes:UpdateCommunicationsStatus()
	local wasInRaid = Hermes.Player.raid
	local wasInParty = Hermes.Player.party
	local wasInBattleground = Hermes.Player.battleground
	
	Hermes:UpdatePlayerGroupStatus()
	
	--this is a special case required so that whenever we go from a party to a raid, or vice verse, that we reset sending and receiving.
	--we can just stop it here if that's the case, and allow the code below to restart it if necessary
	--if((wasInRaid == true and self.PlayerInRaid == false) or (wasInParty == true and self.PlayerInParty == false)) then
	if(	(wasInParty == true and wasInRaid == false and Hermes.Player.raid == true) 			--we just converted from a party to a raid
		or 
		(wasInRaid == true and Hermes.Player.raid == false and Hermes.Player.party == true) ) then	--we just converted from a raid to a party
		
		if(self:IsSending() == true ) then
			self:StopSending()
		end
		if(self:IsReceiving() == true ) then
			self:StopReceiving()
		end
	end

	--start sending if needed
	if( dbp.enabled == true and ((dbp.sender.enabled == true and ((Hermes.Player.party == true and dbp.enableparty == true) or Hermes.Player.raid == true)) or dbp.configMode == true) ) then
		if(self:IsSending() == false) then
			self:StartSending()
		end
	else
		if(self:IsSending() == true ) then
			self:StopSending()
		end
	end
	
	--start receiving if needed
	if(dbp.enabled == true and ((dbp.receiver.enabled == true and ((Hermes.Player.party == true and dbp.enableparty == true) or Hermes.Player.raid == true)) or dbp.configMode == true) ) then
		if(self:IsReceiving() == false) then
			self:StartReceiving()
		end
	else
		if(self:IsReceiving() == true ) then
			self:StopReceiving()
		end
	end
end

function Hermes:RAID_ROSTER_UPDATE()
	self:UpdateCommunicationsStatus()
end

function Hermes:PARTY_MEMBERS_CHANGED()
	self:UpdateCommunicationsStatus()
end

function Hermes:ACTIVE_TALENT_GROUP_CHANGED()
	--reset sender
	Hermes:StopSending()
	Hermes:StartSending()
end

function Hermes:UNIT_SPELLCAST_SUCCEEDED(event, unit, spell, rank, unused, id)
	if unit ~= "player" then return end
	if id == SPELLID_SOULSTONERESURRECTION then
		STARTTIME_SOULSTONERESURRECTION = GetTime() --remember when the spell was cast
	end
end

--------------------------------------------------------------------
-- Receiver
--------------------------------------------------------------------
function Hermes:StartReceiving()
	Hermes.Receiving = true
	
	--give display stuff time to initialize itself before dumping abilities at it
	Hermes:FireEvent("OnStartReceiving")
	
	--initialize Ability list, it should already be empty
	for i, spell in ipairs(dbp.spells) do
		if(spell.enabled and spell.enabled == true) then
			self:StartTrackingAbility(spell)
		end
	end
	
	for i, item in ipairs(dbp.items) do
		if(item.enabled and item.enabled == true) then
			self:StartTrackingAbility(item)
		end
	end
	
	self:StartSenderStatusTimer()

	--be ready for messages before sending requests
	self:RegisterComm(HERMES_RECEIVE_COMM, "OnReceiverComm")

	--advertise receiver
	self:SendMessage_INITIALIZE_RECEIVER(nil)
end

function Hermes:StopReceiving()
	--stop getting messages
	self:UnregisterComm(HERMES_RECEIVE_COMM)
	
	self:StopSenderStatusTimer()
	self:RemoveAllSenders()
	self:StopTrackingAllAbilities()
	
	Hermes.Receiving = false
	
	--initialize display stuff
	Hermes:FireEvent("OnStopReceiving")
end

function Hermes:RequestAbilityUpdate(id)
	local trackerRequests = {}
	tinsert(trackerRequests, id)
	self:SendMessage_REQUEST_SPELLS(nil, trackerRequests)
end

function Hermes:HandleRemoteSender(senderName, classEnum, resetIfExists)
	--see if we know the sender
	local sender = self:FindSenderByName(senderName)
	if(sender) then
		if(not resetIfExists) then
			return
		else
			self:RemoveSender(sender)
		end
	end
	
	if(not sender) then
		self:AddSender(senderName, classEnum)
	end
	
	--build up spellRequests
	local requests = {}
	for _, ability in ipairs(Hermes.Abilities) do
		--only add spells that apply to the class
		if(Hermes:GetClassEnum(ability.class) == classEnum or ability.class == "ANY") then
			tinsert(requests, ability.id)
		end
	end
	
	--only send message if there are spells requested for the senders class
	if(#requests > 0) then
		self:SendMessage_REQUEST_SPELLS(senderName, requests)
	end
end

function Hermes:ProcessMessage_INITIALIZE_SENDER(senderName, classEnum, channel)
	if senderName and senderName == Hermes.Player.name then
		--print("INITIALIZE_SENDER caught from self, ignoring")
	else
		self:HandleRemoteSender(senderName, classEnum, channel ~= "WHISPER")
	end
end

function Hermes:HandleTrackerUpdates(senderName, classEnum, trackerUpdates)
	self:HandleRemoteSender(senderName, classEnum, false) --adds the sender if it wasn't already known
	
	for _, trackerUpdate in ipairs(trackerUpdates) do
		local ability = nil
		for _, a in ipairs(Hermes.Abilities) do
			if(a.id == trackerUpdate[1] and (Hermes:GetClassEnum(a.class) == classEnum or a.class == "ANY")) then
				ability = a --we're tracking it
				break
			end
		end
			
		if(ability) then
			--now update it
			local duration = trackerUpdate[2]

			--find the sender
			local sender = self:FindSenderByName(senderName)
			
			if(duration ~= Hermes.ITEM_NOT_IN_IVENTORY_OR_EQUIPPED) then
				--duration changed and the item is a spell or in the players inventory
				self:SetAbilityInstance(ability, sender, duration)
			else
				--the item is no longer, or never was, in the senders inventory
				local instance = self:FindAbilityInstance(ability, sender)
				if instance then
					self:RemoveAbilityInstance(instance)
				end
			end
		end
	end
end

function Hermes:ProcessMessage_UPDATE_SPELLS(senderName, classEnum, trackerUpdates, channel)
	self:HandleTrackerUpdates(senderName, classEnum, trackerUpdates)
end

--------------------------------------------------------------------
-- Sender
--------------------------------------------------------------------
function Hermes:StartSending()
	--init Sender
	Hermes.Sender = {}
	Hermes.Sender.Trackers = {}

	--initialize display stuff
	Hermes:FireEvent("OnStartSending")
	
	--be ready for messages before start sending
	self:RegisterComm(HERMES_SEND_COMM, "OnSenderComm")

	--register for other wow events
	self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
	
	self:RegisterEvent("ZONE_CHANGED") --handled in Player.lua
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA") --handled in Player.lua
	self:RegisterEvent("ZONE_CHANGED_INDOORS") --handled in Player.lua
		
	--special case for soulstone tracking
	if Hermes.Player.class == self:GetClassEnum("WARLOCK") then
		self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	end

	--add yourself as a sender
	self:AddSender(Hermes.Player.name, Hermes.Player.class)
	
	--advertise self to receivers
	self:SendMessage_INITIALIZE_SENDER(nil)

	if dbp.configMode == false then
		print(string.format("|cFF00FF00Hermes|r: "..L["queuing requests for %s seconds..."], tostring(COOLDOWN_SCAN_FREQUENCY_INITIAL)))
		self:StartCooldownScanTimer(COOLDOWN_SCAN_FREQUENCY_INITIAL)
	else
		self:StartCooldownScanTimer(COOLDOWN_SCAN_FREQUENCY)
	end
end

function Hermes:StopSending()
	--cleanup sender state
	if(Hermes.Sender) then
		wipe(Hermes.Sender)
		Hermes.Sender = nil
	end
	
	--stop getting messages
	self:UnregisterComm(HERMES_SEND_COMM)
	
	--unregister for other wow events
	self:UnregisterEvent("ACTIVE_TALENT_GROUP_CHANGED")
	
	self:UnregisterEvent("ZONE_CHANGED")
	self:UnregisterEvent("ZONE_CHANGED_NEW_AREA")
	self:UnregisterEvent("ZONE_CHANGED_INDOORS")
		
	self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED") --special case for Soulstone tracking improvements
	self:KillCooldownScanTimer()
	
	--initialize display stuff
	Hermes:FireEvent("OnStopSending")
end

function Hermes:HandleTrackerRequests(receiverName, trackerRequests)
	for _, tid in ipairs(trackerRequests) do
		if(Hermes:SenderHasTracker(tid)) then
			local tracker = nil
			--see if we know the tracker
			for _, t in ipairs(Hermes.Sender.Trackers) do
				if(t.id == tid) then
					--we know the tracker, see if we know the receiver
					local receiver = t.receivers[receiverName]
					--we don't know this receiver yet, add it
					if(not receiver) then
						t.receivers[receiverName] = {}
					end
					
					--mark this receiver as dirty
					if(not t.dirtyReceivers) then
						t.dirtyReceivers = {}
					end
					tinsert(t.dirtyReceivers, receiverName)
					
					tracker = t --mark spell for below...
					break
				end
			end
			
			--we're not tracking this spell yet
			if(not tracker) then
				tracker = {
					id = tid,
					duration = nil,
					receivers = {}
				}
				tracker.receivers[receiverName] = {} --add the receiver to the spell
				tinsert(Hermes.Sender.Trackers, tracker) --add the spell
				
				--mark this receiver as dirty
				tracker.dirtyReceivers = {}
				tinsert(tracker.dirtyReceivers, receiverName)
			end
		end
	end
end

function Hermes:ProcessMessage_REQUEST_SPELLS(receiverName, TrackerRequests, channel)
	self:HandleTrackerRequests(receiverName, TrackerRequests)
end

function Hermes:ProcessMessage_INITIALIZE_RECEIVER(receiverName, channel)
	if receiverName and receiverName == Hermes.Player.name then
		--print("INITIALIZE_RECEIVER caught from self, ignoring")
	else
		self:SendMessage_INITIALIZE_SENDER(receiverName)
	end
end

function Hermes:IsSending()
	return Hermes.Sender ~= nil
end

function Hermes:IsReceiving()
	return Hermes.Receiving
end

--------------------------------------------------------------------
-- AceAddon
--------------------------------------------------------------------
function Hermes:OnEnable()
	self:Startup()
end

function Hermes:OnDisable()
	self:Shutdown()
end

function Hermes:Startup() --called when hermes is enabled
	--reset player status so that it detects as changed
	Hermes.Player.raid = false
	Hermes.Player.party = false
	Hermes.Player.battleground = false
	
	--create default spells and items for new profile
	if(not dbp.welcome)	then
		dbp.welcome = true
		Hermes:SetupNewProfileButtons()
	end

	--enable all registered plugins so that they can start hooking events
	self:EnablePlugins()
	
	--update the entire option tables, we have to call this after enabling plugins so they they can setup their option tables
	self:UpdateBlizOptionsTableReferences()
		
	--start Sending and Receiving as needed
	if(PLAYER_ENTERED_WORLD) then
		--go right into config mode if so configured, make sure to do this before calling UpdateCommunicationsStatus so that we go into test mode before any other mode.
		if(dbp.configMode == true) then
			self:StartTestMode()
		else
			self:UpdateCommunicationsStatus()
		end
	else
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
	end

	self:RegisterEvent("RAID_ROSTER_UPDATE")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	
	--fire off any item name timers
	for _, item in ipairs(dbp.items) do
		if(item.caching) then
			Hermes:StartItemNameTimer(item.id) --fire off timer for getting item names
		end
	end
end

function Hermes:Shutdown() --called when hermes is disabled
	self:UnregisterEvent("RAID_ROSTER_UPDATE")
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED")
	self:StopSending()
	self:StopReceiving()
	
	--disable plugins before we stop receiving, very important because they may need to unhook events
	self:DisablePlugins()
	
	self:KillItemNameTimer()
end

function Hermes:OnInitialize() -- Only called once, do all one time init stuff here
	self:InitializeHermesPluginTable()
	self:InitializeHermesEvents()
	
	--load the absolute bare essentials for options. These merge with options saved already
	self:LoadDefaultOptions()
	
	--create a reference to the profile table
	dbp = self.db.profile

	--initialize option tables
	self:LoadBlizOptions()
	
	--create the frame which uses OnUpdate event to manage spell durations and OnUpdate events
	self:InitializeAbilityInstanceFrame()
	
	--setup profile events
	self.db.RegisterCallback(self, "OnNewProfile", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
end

function Hermes:SetupNewProfileSpells()
	local playerAlliance = select(1, UnitFactionGroup("player"))
	
	for i, default in ipairs(Hermes.DEFAULT_SPELLS) do
		local class = default[1]
		local spellid = default[2]
		local alliance = default[3]
		
		if(not alliance or (alliance and alliance == playerAlliance)) then
			--make sure spell isn't already added
			local exists = false
			for _, s in ipairs(dbp.spells) do
				if(s.id == spellid) then
					exists = true
					break
				end
			end
			
			if(exists == false) then
				local spell = Hermes:FindSpellName(class, spellid)
				if(spell) then
					--add to db
					spell.enabled = false
					tinsert(dbp.spells, spell)
					--sort the spells
					table.sort(dbp.spells, function(a, b) return Hermes:SortProfileSpells(a, b) end)
				end
			end
		end
	end
	
	--update blizzard options
	Hermes:BlizOptionsTable_Spells()
end

function Hermes:SetupNewProfileItems()
	local playerAlliance = select(1, UnitFactionGroup("player"))
	
	for i, default in ipairs(Hermes.DEFAULT_ITEMS) do
		local class = default[1]
		local id = default[2]
		local alliance = default[3]
		
		if(not alliance or (alliance and alliance == playerAlliance)) then
			--make sure item isn't already added
			local exists = false
			for _, s in ipairs(dbp.items) do
				if(s.id == self:EncodeTrackerIdentifier(id, "item")) then
					exists = true
					break
				end
			end
			
			if(exists == false) then
				local itemid, itemname, itemicon = self:GetItemInfo(id, nil)
				
				if(itemicon and itemid) then
					itemid = self:EncodeTrackerIdentifier(itemid, "item")
					--we may need to cache the item
					local caching = nil
					if(not itemname) then
						caching = 1
						itemname = "..."..L["searching"].." ("..tostring(self:DecodeTrackerIdentifier(tonumber(itemid)))..")"
					end
				
					local item = {
						class = class,
						id = itemid,
						name = itemname,
						icon = itemicon,
						enabled = false,
						caching = caching
					}
				
					--add to db
					tinsert(dbp.items, item)
				
					--sort the items
					table.sort(dbp.items, function(a, b) return Hermes:SortProfileItems(a, b) end)
				
					--fire off the name cache timer if needed
					if(caching) then
						Hermes:StartItemNameTimer(item.id)
					end
				end
			end
		end
	end
	
	--update blizzard options
	Hermes:BlizOptionsTable_Items()
end

function Hermes:SetupNewProfileButtons()
	self:SetupNewProfileSpells()
	self:SetupNewProfileItems()
end

function Hermes:FindSpellName(class, spellid)
	--see if the spell exists
	local name
	local icon
	name, _, icon, _, _, _, _, _, _ = GetSpellInfo(spellid)
	
	if(not name or not icon) then
		return nil
	end
	
	local spell = {
		class = class,
		id = spellid,
		name = name,
		icon = icon,
		enabled = false,
	}

	return spell
end

function Hermes:OnProfileChanged()
	dbp = self.db.profile
	self:Disable()
	
	--if they switched to an different profile since upgrading Hermes
	--self:DoDatabaseUpgrade()
	
	self:UpdateBlizOptionsTableReferences() --update table references, otherwise they'll be pointing to old tables for the prior profile

	self:Enable()
	
	if dbp.configMode == true then
		self:ShowHermesTestModeMessage()
	end
end

--------------------------------------------------------------------
-- Message Sending
--------------------------------------------------------------------
function Hermes:SendMessageToReceivers(recipientName, msgEnum, msgContent)
	local msg = self:Serialize({msgEnum, msgContent})
	local channel, recipient = self:GetAppropriateMessageChannelAndRecipient(recipientName)
	--print("Sender to Receiver: "..channel..","..tostring(recipient)..","..tostring(MESSAGE_ENUM[msgEnum]))
	self:SendCommMessage(HERMES_RECEIVE_COMM, msg, channel, recipient, "NORMAL", nil, nil)
end

function Hermes:GetAppropriateMessageChannelAndRecipient(recipientName)
	if(dbp.configMode == true) then
		return "WHISPER", Hermes.Player.name --in test mode, only send whispers to yourself
	elseif(recipientName) then
		return "WHISPER", recipientName
	elseif(Hermes.Player.battleground == true) then
		return "BATTLEGROUND", recipientName
	elseif(Hermes.Player.raid == true) then
		return "RAID", recipientName
	elseif(Hermes.Player.party == true) then
		return "PARTY", recipientName
	else
		error("Unable to determine message channel")
	end
end

function Hermes:SendMessageToSenders(recipientName, msgEnum, msgContent)
	local msg = self:Serialize({msgEnum, msgContent})
	local channel, recipient = self:GetAppropriateMessageChannelAndRecipient(recipientName)
	--print("Receiver to Sender: "..channel..","..tostring(recipient)..","..tostring(MESSAGE_ENUM[msgEnum]))
	self:SendCommMessage(HERMES_SEND_COMM, msg, channel, recipient, "NORMAL", nil, nil)
end

function Hermes:SendMessage_REQUEST_SPELLS(recipientName, trackerRequests)
	self:SendMessageToSenders(recipientName, Hermes:GetMessageEnum("REQUEST_SPELLS"),
		{
			trackerRequests
		}
	)
end

function Hermes:SendMessage_INITIALIZE_SENDER(recipientName)
	self:SendMessageToReceivers(recipientName, Hermes:GetMessageEnum("INITIALIZE_SENDER"),
		Hermes.Player.class
	)
end

function Hermes:SendMessage_INITIALIZE_RECEIVER(recipientName)
	self:SendMessageToSenders(recipientName, Hermes:GetMessageEnum("INITIALIZE_RECEIVER"),
		nil
	)
end

function Hermes:SendMessage_UPDATE_SPELLS(recipientName, trackerUpdates)
	self:SendMessageToReceivers(recipientName, Hermes:GetMessageEnum("UPDATE_SPELLS"),
		{
			Hermes.Player.class,
			trackerUpdates,
		}
	)
end

--------------------------------------------------------------------
-- Message Receiving
--------------------------------------------------------------------
function Hermes:OnReceiverComm(prefix, serialized, channel, sender)
	if(prefix == HERMES_RECEIVE_COMM) then
	
		--when toggling config mode, it's possible to get stale messages from players after having already set config mode.
		--So if in config mode, make sure that the message if a message from yourself, sent to the whisper channel
		if dbp.configMode  == true and channel ~= "WHISPER" and sender ~= Hermes.Player.name then
			--ignore the message
			return
		end
		
		local success, msg = self:Deserialize(serialized)
		if(success) then
			local msgEnum = msg[1]
			local msgContent = msg[2]
			local msgName = MESSAGE_ENUM[msgEnum]
			
			--print("Message from Sender: "..tostring(sender)..", "..tostring(channel)..", "..tostring(msgName))
			
			if(msgName and msgName == "INITIALIZE_SENDER") then
				local classEnum = msgContent
				self:ProcessMessage_INITIALIZE_SENDER(sender, classEnum, channel)
				
			elseif(msgName and msgName == "UPDATE_SPELLS") then
				local classEnum = msgContent[1]
				local trackerUpdates = msgContent[2]
				self:ProcessMessage_UPDATE_SPELLS(sender, classEnum, trackerUpdates, channel)
			end
		else
			error("Error deserializing message")
		end
	end
end

function Hermes:OnSenderComm(prefix, serialized, channel, sender)
	if(prefix == HERMES_SEND_COMM) then
		--when toggling config mode, it's possible to get stale messages from players after having already set config mode.
		--So if in config mode, make sure that the message if a message from yourself, sent to the whisper channel
		if dbp.configMode == true and channel ~= "WHISPER" and sender ~= Hermes.Player.name then
			--ignore the message
			return
		end
		
		local success, msg = self:Deserialize(serialized)
		if(success) then
			local msgEnum = msg[1]
			local msgContent = msg[2]
			local msgName = MESSAGE_ENUM[msgEnum]

			--print("Message from Receiver: "..tostring(sender)..", "..tostring(channel)..", "..tostring(msgName))
			
			if(msgName and msgName == "REQUEST_SPELLS") then
				local trackerRequests = msgContent[1]
				self:ProcessMessage_REQUEST_SPELLS(sender, trackerRequests, channel)

			elseif(msgName and msgName == "INITIALIZE_RECEIVER") then
				self:ProcessMessage_INITIALIZE_RECEIVER(sender, channel)
			end
		else
			error("Error deserializing message")
		end
	end
end

--------------------------------------------------------------------
-- Spell Tracking
--------------------------------------------------------------------
function Hermes:SenderHasTracker(identifier)
	local id, idtype = self:DecodeTrackerIdentifier(identifier)
	
	if(idtype == "spell") then
		return self:IsSpellKnown(id)
	else
		--always return trueThis is because for all we know they might have changed gear, gone to the bank, etc.
		--TODO: On the next version of the protocol, make sure that the receiver sends the class of the spell when talking to senders
		--return self:IsItemKnown(id)
		return true
	end
end

function Hermes:IsSpellKnown(spellid)
	local slot = FindSpellBookSlotBySpellID(spellid)
	if slot and slot <= MAX_SPELLS then
		local slotType, id = GetSpellBookItemInfo(slot, BOOKTYPE_SPELL)
		return id == spellid and slotType ~= 'FUTURESPELL'
	end
end

local equippableSlots = {
	"AmmoSlot",
	"BackSlot",
	"ChestSlot",
	"FeetSlot",
	"Finger0Slot",
	"Finger1Slot",
	"HandsSlot",
	"HeadSlot",
	"LegsSlot",
	"MainHandSlot",
	"NeckSlot",
	"RangedSlot",
	"SecondaryHandSlot",
	"ShirtSlot",
	"ShoulderSlot",
	"TabardSlot",
	"Trinket0Slot",
	"Trinket1Slot",
	"WaistSlot",
	"WristSlot",
}

function Hermes:IsItemKnown(itemid)
	--50620 Coldwraith Links
	--50364 Sindragosa's Flawless Fang
	
	if(IsEquippableItem(itemid)) then --figure out whether it's something that's supposed to be equipped or which sits in inventory
		local itemCount = GetItemCount(itemid, nil, nil) --see if they have the item at all (not including bank)
		if(itemCount > 0) then --the person has the item, see if it's equipped
			for _, slot in ipairs(equippableSlots) do
				local slotid = GetInventorySlotInfo(slot)
				local inventoryid = GetInventoryItemID("player", slotid)
				if(inventoryid == itemid) then --item found
					return true
				end
			end
		end
	else
		local itemCount = GetItemCount(itemid, nil, nil) --see if they have the item
		if(itemCount > 0) then --the person has the item, no need to check for equipped
			return true
		end
	end
	
	--item not found
	return false
end

function Hermes:UpdateSenderCooldown(tracker)
	local send = false

	--get latest cooldown info
	local priorCooldown = tracker.duration
	tracker.duration = self:GetLatestCooldown(tracker.id)
		
	--[[
	If the spell being tracked is determined to be one which uses runes and which also has an actual cooldown,
	then set tracker.duration and priorCooldown to nil. This fools Hermes into thinking that the cooldown
	has ended "naturally" and that no messages will need to be sent as updates, or in the case that it was forced above, it will
	be sent as being available.
	]]--
	if(Hermes.Player.class == self:GetClassEnum("DEATHKNIGHT")) then
		if(self:AdjustForRunes(tracker)) then
			tracker.duration = nil
			priorCooldown = nil
		end
	end	

	--special handling for warlock soulstones
	if(Hermes.Player.class == self:GetClassEnum("WARLOCK")) then
		--see if this is the soulstone tracker
		if tracker.id == ITEMID_SOULSTONE then
			local maxDuration = 15 * 60 --15 minutes
			if tracker.duration == Hermes.ITEM_NOT_IN_IVENTORY_OR_EQUIPPED then
				--no soulstone in inventory, let's try looking at STARTTIME_SOULSTONERESURRECTION instead
				if STARTTIME_SOULSTONERESURRECTION ~= nil then
					--UNIT_SPELLCAST_SUCCEEDED was caught at some point, let's see if it's expired yet
					local duration = ceil(STARTTIME_SOULSTONERESURRECTION - (GetTime() - maxDuration))
					
					if duration < 0 then
						--we're here because two things have happened:
						--1) The user has no soulstone in their inventory
						--2) The soulstone resurrection timer was running, but it just expired
						--Net result is that the spell is known to be available now
						tracker.duration = nil
						
						--reset the resurrection time
						STARTTIME_SOULSTONERESURRECTION = nil
					else
						--we're here because two things have happened:
						--1) The user has no soulstone in their inventory
						--2) The soulstone resurrection timer is still running
						--Net result is that the spell is on cooldown
						tracker.duration = duration
					end
				else
					--Both Soulstone item and Soulstone Resurrection times are nil.
					--We have no idea if the spell is actually available or not so just assume it's available
					--As soon as the warlock creates a soulstone it'll update so this is the best bet.
					tracker.duration = nil
				end
			else
				if tracker.duration ~= nil then
				--We're here because there is in fact a soulstone in their inventory
				--Let's go ahead and set STARTTIME_SOULSTONERESURRECTION to an appropriate value
					STARTTIME_SOULSTONERESURRECTION = GetTime() + (tracker.duration - maxDuration)
				else
					--soulstone in inventory and the spell is available, don't need to do anything as tracker.duration will already be correct (nil)
				end
			end
		end
	end
	
	--determine whether we need to send any global messages
	if(priorCooldown ~= tracker.duration) then
		--if it wasn't on cooldown last but it is now
		if(priorCooldown == nil and tracker.duration ~= nil) then
			send = true

			--if it was on cooldown but now it's not.
		elseif(priorCooldown ~= nil and tracker.duration == nil) then
			--only send message if the last cooldown value is greater than the threshold, this is to prevent sending messages for spells that went off cooldown naturally
			if(priorCooldown == Hermes.ITEM_NOT_IN_IVENTORY_OR_EQUIPPED or priorCooldown > COOLDOWN_DELTA_THRESHOLD) then
				send = true
			end
				
		elseif(priorCooldown ~= nil and tracker.duration ~= nil) then
			--if for some reason the cooldown is larger than it was last time we checked
			if(tracker.duration > priorCooldown) then
				send = true
			end
			--if for some reason the cooldown is unexpectedly smaller than last time we checked
			if((priorCooldown - tracker.duration) > COOLDOWN_DELTA_THRESHOLD) then
				send = true
			end
		end
	end

	return send
end

function Hermes:GetLatestCooldown(identifier)
	local id, idtype = Hermes:DecodeTrackerIdentifier(identifier)
	local cooldown = nil
	if(idtype == "spell") then
		local start, duration, _ = GetSpellCooldown(id)
		--note that GetSpellCooldown returns the GCD if active when called. By calling > 2 I hope to filter out any GCD's at the expense of not reporting cooldowns actually only having 2 seconds or less on them.
		if(duration > 2) then
			cooldown = ceil(duration - (GetTime() - start)) --spell on cooldown
		else
			cooldown = nil
		end
	else
		if(IsEquippableItem(id)) then --figure out whether it's something that's supposed to be equipped or which sits in inventory
			local itemCount = GetItemCount(id, nil, nil) --see if they have the item at all (not including bank)
			if(itemCount > 0) then --the person has the item, see if it's equipped
				for _, slot in ipairs(equippableSlots) do
					local slotid = GetInventorySlotInfo(slot)
					local inventoryid = GetInventoryItemID("player", slotid)
					if(inventoryid == id) then --item found
						local start, duration, enable = GetInventoryItemCooldown("player", slotid)
						if(duration > 2) then
							cooldown = ceil(duration - (GetTime() - start)) --spell on cooldown
						else
							cooldown = nil
						end
					end
				end
			else
				cooldown = Hermes.ITEM_NOT_IN_IVENTORY_OR_EQUIPPED
			end
		else
			local itemCount = GetItemCount(id, nil, nil) --see if they have the item
			if(itemCount > 0) then --the person has the item, no need to check for equipped
				local start, duration, enable = GetItemCooldown(id)
				if(duration > 2) then
					cooldown = ceil(duration - (GetTime() - start)) --spell on cooldown
				else
					cooldown = nil
				end
				
			else
				cooldown = Hermes.ITEM_NOT_IN_IVENTORY_OR_EQUIPPED
			end
		end
	end

	return cooldown
end

--------------------------------------------------------------------
-- Helpers
--------------------------------------------------------------------
--sees if the player has item information such as id, name, and icon in their cache
function Hermes:GetItemInfoFromPlayerCache(itemId, itemName)
	local id, name, icon
	
	if(itemId) then
		name = GetItemInfo(itemId)
		id = itemId
		icon = GetItemIcon(itemId)
		
	elseif(itemName) then
		name, link = GetItemInfo(itemName)
		if(name) then
			--borrowed expressions from ItemId, thank you
			local justItemId = string.gsub(link, ".-\124H([^\124]*)\124h.*", "%1");
			local type, itemid, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId = strsplit(":",justItemId);
			id = tonumber(itemid) --todo, extract id from link
			icon = GetItemIcon(id)
		end
	end
	
	return id, name, icon
end

--tries to find an item's game information by name or id, choosing id first if specified
function Hermes:GetItemInfo(requestid, requestname)
	local id, name, icon
	if(requestid) then
		--this method will always return a value if the id exists
		icon = GetItemIcon(requestid)
		--the item by that id exists, but it's not in the local or server cache, try to query again later.
		if(icon) then --try to lookup item again
			--item exists, see if we have the name in cache, could return nil
			id, name, _ = self:GetItemInfoFromPlayerCache(requestid, nil)
		end
	else
		id, name, icon = self:GetItemInfoFromPlayerCache(nil, requestname)
	end
	
	return id, name, icon
end

function Hermes:GetClassEnum(classFileName)
	for i, class in ipairs(Hermes.CLASS_ENUM) do
		if(class == classFileName) then
			return i
		end
	end
	error("Failed to locate class enum")
end

function Hermes:GetMessageEnum(messageName)
	for i, msg in ipairs(MESSAGE_ENUM) do
		if(msg == messageName) then
			return i
		end
	end
	error("Failed to locate message enum")
end

function Hermes:KillCooldownScanTimer()
	if(CooldownScanTimer) then
		self:CancelTimer(CooldownScanTimer, true)
		CooldownScanTimer = self:ScheduleTimer("OnUpdateSenderCooldowns", 0.1, nil) --convert to one shot timer
		self:CancelTimer(CooldownScanTimer, true) --cancel again, we don't care if it ran or not
		CooldownScanTimer = nil
	end
end

function Hermes:StartCooldownScanTimer(delay)
	CooldownScanTimer = self:ScheduleRepeatingTimer("OnUpdateSenderCooldowns", delay, delay)
end

function Hermes:OnUpdateSenderCooldowns(delay)
	if(Hermes.Sender and Hermes.Sender.Trackers) then
		--first update all of the cooldowns and populate list with spells that need to be sent
		local trackerUpdates = nil
		for i, tracker in ipairs(Hermes.Sender.Trackers) do
			--if...
			--	1. There are dirty receivers
			--	2. Dirty receiver count is equal to total sender count.
			--	3. Dirty receiver count is greater than zero.
			--	4. Message needs to be sent due to cooldown change...
			-- The point of this is to avoid sending a message to everyone when there is only one receiver interested in the spell, in which case a single message will be sent to the user
			if(self:UpdateSenderCooldown(tracker) == true or ((	tracker.dirtyReceivers and Hermes:GetTableCount(tracker.receivers) == #tracker.dirtyReceivers) and (tracker.dirtyReceivers and #tracker.dirtyReceivers > 1) )) then
				if(not trackerUpdates) then
					trackerUpdates = {}
				end
				tinsert(trackerUpdates, {tracker.id, tracker.duration})
				
				--wipe out dirty receivers since a global send is going to be done
				tracker.dirtyReceivers = nil
			end
		end
		
		--There is more than one receiver interested in this spell that also requires an update, fire it off globally
		if(trackerUpdates and #trackerUpdates > 0) then
			self:SendMessage_UPDATE_SPELLS(nil, trackerUpdates)
		end
		
		--cleanup table
		local trackerUpdates = nil
		
		local dirtyUpdates = nil
		--loop again looking for any trackers that have dirty receivers, if a global message was sent above then this won't find anything
		for i, tracker in ipairs(Hermes.Sender.Trackers) do
			if(tracker.dirtyReceivers) then
				for _, receiverName in ipairs(tracker.dirtyReceivers) do
					--create dirty table if not already exists
					if(not dirtyUpdates) then dirtyUpdates = {}	end
				
					--add receiver if not already added
					if(not dirtyUpdates[receiverName]) then
						dirtyUpdates[receiverName] = {}
					end
				
					--add the update to the receiver record
					tinsert(dirtyUpdates[receiverName], {tracker.id, tracker.duration})
				end
				
				--wipe out dirty receivers for this tracker
				tracker.dirtyReceivers = nil
			end
		end
		
		-- these individual receivers still need an update, a message is sent to each instead of spammed to everyone.
		-- the performance boost here is that the sender and a few receivers take the hit instead of all the receivers (perhaps 25 of them!)
		if(dirtyUpdates) then
			for receiverName, trackerUpdates in pairs(dirtyUpdates) do
				self:SendMessage_UPDATE_SPELLS(receiverName, trackerUpdates)
			end
		end
		
		--wipe out table if exists
		dirtyUpdates = nil
	end
	
	--if this was the initial scan from just starting up, then restart the timer with the regular scan rate
	if(delay == COOLDOWN_SCAN_FREQUENCY_INITIAL) then
		self:KillCooldownScanTimer()
		self:StartCooldownScanTimer(COOLDOWN_SCAN_FREQUENCY)
		print("|cFF00FF00Hermes|r: "..L["now sending"])
	end
end

function Hermes:GetLocalizedClassName(classToken)
	local localizedClassName
	if(classToken == "ANY") then
		return L["Any"]
	else
		return localizedClassNames[classToken]
	end
end

function Hermes:GetClassColorRGB(class)
	local number = tonumber(class)
	if(number) then --they used the enum number
		local classColorRGB
		if(class == 1) then
			classColorRGB = RAID_CLASS_COLORS["PRIEST"] --white
		else
			classColorRGB = RAID_CLASS_COLORS[Hermes.CLASS_ENUM[class]]
		end
		return classColorRGB
	else --they used the enum value
		local classColorRGB
		if(class == "ANY") then
			classColorRGB = RAID_CLASS_COLORS["PRIEST"] --white
		else
			classColorRGB = RAID_CLASS_COLORS[class]
		end
		return classColorRGB
	end
end

function Hermes:GetClassColorHEX(class)
	local number = tonumber(class)
	if(number) then --they used the enum number
		local classColorRGB
		if(class == 1) then
			classColorRGB = RAID_CLASS_COLORS["PRIEST"] --white
		else
			classColorRGB = RAID_CLASS_COLORS[Hermes.CLASS_ENUM[class]]
		end
		local classColorHex = string.format("FF%02x%02x%02x", 
			classColorRGB.r * 255,
			classColorRGB.g * 255,
			classColorRGB.b * 255)
		return classColorHex
	else --they used the enum value
		local classColorRGB
		if(class == "ANY") then
			classColorRGB = RAID_CLASS_COLORS["PRIEST"] --white
		else
			classColorRGB = RAID_CLASS_COLORS[class]
		end
		local classColorHex = string.format("FF%02x%02x%02x", 
			classColorRGB.r * 255,
			classColorRGB.g * 255,
			classColorRGB.b * 255)
		return classColorHex
	end
end

function Hermes:GetColorCodedClassText(text, class)
	return "|c"..self:GetClassColorHEX(class)..text.."|r"
end

function Hermes:Trim (s)
	return (string.gsub(s, "^%s*(.-)%s*$", "%1"))
end

function Hermes:SortProfileSpells(a, b)
	return a.name < b.name
end

function Hermes:SortProfileItems(a, b)
	return a.name < b.name
end

function Hermes:GetSpellID(spell)
	local matches = {}
	for i = 1, 100000 do
		name, rank, icon, powerCost, _, powerType, castingTime, minRange, maxRange = GetSpellInfo(i)
		if name == spell then
			tinsert(matches, i)
		end
	end
	
	return matches
end

function Hermes:DecodeTrackerIdentifier(id)	--spells have positive id's, items have negative id's, returns the type and a positive id
	if(id >= 0) then
		return id, "spell"
	elseif(id < 0) then
		return math.abs(id), "item"
	end
end

function Hermes:EncodeTrackerIdentifier(id, trackertype)	--spells have positive id's, items have negative id's, returns the type and a positive id
	if(trackertype == "spell") then
		return math.abs(id)
	elseif(trackertype == "item") then
		return math.abs(id) * -1
	else
		error("unknown encode type")
	end
end

--------------------------------------------------------------------
-- Item Name Timer Cache Check Methods
--------------------------------------------------------------------
function Hermes:GetNextItemIdToCache(currentId)
	local lastIndex = nil
	local nextIndex = nil
	for i, item in ipairs(dbp.items) do
		if(lastIndex and item.caching) then
			nextIndex = i
			break
		end
		if(item.id == currentId) then
			lastIndex = i
		end
	end
	
	--it's possible there weren't any items needing caching ocurring after the last item,
	--so start from beginning and take the first match. This will also restart the current item if it's the first one
	if(lastIndex and not nextIndex) then
		for i, item in ipairs(dbp.items) do
			if(item.caching) then
				nextIndex = i
				break
			end
		end
	end
	
	if(nextIndex) then
		return dbp.items[nextIndex].id
	end
end

function Hermes:OnUpdateItemNameTimer(currentId)
	Hermes:KillItemNameTimer() 
	--see if the client knows the name now
	local id, name, icon = self:GetItemInfoFromPlayerCache(self:DecodeTrackerIdentifier(currentId), nil)
	if(id and name and icon) then
		--client knows the item now, update it
		for _, item in ipairs(dbp.items) do
			if(item.id == currentId) then
				item.caching = nil
				item.name = name
				item.icon = icon
				--update hermes display
				--local _, button = HermesDisplay:GetContainerAndCellForTrackerId(currentId)
				--it might not be showing right now
				--if(button) then
				--	button.trackerInfo.name = item.name
				--end
			end
		end
		
		--resort the items to account for new name
		table.sort(self.db.profile.items, function(a, b) return Hermes:SortProfileItems(a, b) end)
		
		--update blizz options table with the new info
		LIB_AceConfigRegistry:NotifyChange(AddonName)
		Hermes:BlizOptionsTable_Items()
	end
	
	--now let's see if we need to kick off another timer
	local nextId = self:GetNextItemIdToCache(currentId)
	if(nextId) then
		ItemNameTimer = self:ScheduleTimer("OnUpdateItemNameTimer", ITEM_NAME_THROTTLE, nextId)
	end
end

function Hermes:KillItemNameTimer()
	if(ItemNameTimer) then
		self:CancelTimer(ItemNameTimer, true)
		ItemNameTimer = nil
	end
end

function Hermes:StartItemNameTimer(id)
	--don't kick off the timer if it's already running, it's smart enough to restart itself
	if(ItemNameTimer == nil) then
		--if it wasn't already running, then make the first one 10 seconds for quicker initial response of a new item
		ItemNameTimer = self:ScheduleTimer("OnUpdateItemNameTimer", 5, id)
	end
end

--------------------------------------------------------------------
-- Special Handling for DK Rune Cooldowns
-- Only spells using runes that actually have a cooldown need to be listed
-- This algorithm will only work properly so long as all cooldowns using
-- runes are greater than the rune duration of 10 (or 9 depending on talents) seconds.
--------------------------------------------------------------------
local runeSpells = {
	[51052] = 0, -- Anti-Magic Zone -- 2 min
	[42650] = 0, -- Army of the Dead -- 10 min
	[49222] = 0, -- Bone Shield -- 1 min
	[43265] = 0, -- Death and Decay -- 30 sec
	[50977] = 0, -- Death Gate 1 min
	[51271] = 0, -- Pillar of Frost -- 1 min
	[48982] = 0, -- Rune Tap -- 30 sec
	[47476] = 0, -- Strangulate -- 2 min
}

function Hermes:AdjustForRunes(tracker)
	local runeSpell = runeSpells[tracker.id]
	if (runeSpell and tracker.duration and tracker.duration <= 10) then --10 seconds is good enough. Checking for talents and 9 is not worth the trouble
		return true
	end
	
	return false
end

------------------------------------------------------------------
-- GENERAL HELPERS
------------------------------------------------------------------
function Hermes:FindTableIndex(tbl, item)
	for index, i in ipairs(tbl) do
		if(i == item) then
			return index
		end
	end

	return nil
end

function Hermes:MoveIndexedCell(tbl, item, newIndex)
	if not newIndex then error("nil index") end
	local oldIndex = Hermes:FindTableIndex(tbl, item)
	if not oldIndex then error("failed to locate item") end
	if oldIndex == newIndex then return end
	
	--remove from old location
	tremove(tbl, oldIndex)
	
	--add to new location
	tinsert(tbl, newIndex, item)
end

function Hermes:DeleteFromIndexedTable(tbl, item)
	local index = Hermes:FindTableIndex(tbl, item) 
	if not index then error("failed to locate item in table") end
	tremove(tbl, index)
end

function Hermes:deepcopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end

function Hermes:RotateTexture(self, angle)
	local function GetCorner(angle)
		local Root2 = 2 ^ 0.5;
		return 0.5 + cos( angle ) / Root2, 0.5 + sin( angle ) / Root2;
	end

	local LRx, LRy = GetCorner( angle + 45 );
	local LLx, LLy = GetCorner( angle + 135 );
	local ULx, ULy = GetCorner( angle - 135 );
	local URx, URy = GetCorner( angle - 45 );
	self:SetTexCoord( ULx, ULy, LLx, LLy, URx, URy, LRx, LRy );
end

function Hermes:Round(num, idp)
	local mult = 10^(idp or 0)
	return math.floor(num * mult + 0.5) / mult
end

function Hermes:SecondsToClock(sSeconds)
	local nSeconds = tonumber(sSeconds)
	if nSeconds == 0 then
		return "0";
	else
		nHours = string.format("%02.f", math.floor(nSeconds/3600));
		nMins = string.format("%02.f", math.floor(nSeconds/60 - (nHours*60)));
		nSecs = string.format("%02.f", math.floor(nSeconds - nHours*3600 - nMins *60));
		
		if(nSeconds > 3600) then
			return nHours..":"..nMins..":"..nSecs
		end
		
		if(nSeconds > 60) then
			return nMins..":"..nSecs
		end
		
		return Hermes:Round(nSeconds)
	end
end

function Hermes:GetTableCount(tbl)
	local cnt = 0
	for k, v in pairs(tbl) do
		cnt = cnt + 1
	end
	return cnt
end

function Hermes:TableMerge(t1, t2)
    for k,v in pairs(t2) do
        if type(v) == "table" then
                if type(t1[k] or false) == "table" then
                        Hermes:TableMerge(t1[k] or {}, t2[k] or {})
                else
                        t1[k] = v
                end
        else
                t1[k] = v
        end
    end
    return t1
end

------------------------------------------------------------------
-- CONFIG MODE: Simulate addon running
------------------------------------------------------------------
function Hermes:ShowHermesTestModeMessage()
	local message = "|cff19FF19Hermes: "..L["Config Mode"].."|r\n\n"
	message = message ..L["Hermes is running in Config Mode."].."\r\n"..L["Toggle it on or off in the 'General' settings tab."].."\n\n"
	StaticPopupDialogs["HermesConfigMode"] = {
		text = message, 
		button1 = L["Edit Settings"],
		button2 = L["Close"],
		timeout = 0,
		whileDead = 1, 
		hideOnEscape = 1, 
		OnAccept = function() self:OpenConfigWindow() end,
		OnCancel = function() end
	}
	StaticPopup_Show("HermesConfigMode")
	dbp.welcome = true
end

function Hermes:StartTestMode()
	--set the test mode bit
	dbp.configMode = true
	
	--we may already be a raid or party or some such, stop sending and receiving if we are so they we start with a clean slate
	if self:IsSending() == true then
		self:StopSending()
	end
	if self:IsReceiving() == true  then
		self:StopReceiving()
	end
	
	--force sending and receiving to start in test mode
	self:UpdateCommunicationsStatus()
end

function Hermes:StopTestMode()
	--set the test mode bit
	dbp.configMode = false
	
	--make sure we completely reset the sending/receiving state
	self:StopSending()
	self:StopReceiving()
	
	--force sending and receiving to start (if applicable)
	self:UpdateCommunicationsStatus()
end
