local AddonName, Hermes = ...

local LIB_AceConfig = LibStub("AceConfig-3.0") or error("Required library AceConfig-3.0 not found")
local LIB_AceDB = LibStub("AceDB-3.0") or error("Required library AceDB-3.0 not found")
local LIB_AceConfigDialog = LibStub("AceConfigDialog-3.0")  or error("Required library AceConfigDialog-3.0 not found")
local LIB_AceConfigRegistry = LibStub("AceConfigRegistry-3.0") or error("Required library AceConfigRegistry-3.0 not found")
local L = LibStub('AceLocale-3.0'):GetLocale('Hermes')
local LIB_LibSharedMedia = LibStub("LibSharedMedia-3.0") or error("Required library LibSharedMedia-3.0 not found")

local optionTables = nil

local newSpellTemplate = {
	class = L["-- Select --"],
	id = nil,
	name = nil,
}

local newItemTemplate = {
	class = L["-- Select --"],
	id = nil,
	name = nil,
}

function Hermes:OpenConfigWindow()
	LIB_AceConfigDialog:Open(AddonName)
end

local slashCommands = {
	name = "Hermes Command Line",
	handler = Hermes,
	type = 'group',
	args = {
		config = {
			type = "execute",
			name = L["Open Configuration"],
			order = 0,
			func = function(info) Hermes:OpenConfigWindow() end,
		},
		enable = {
			type = "toggle",
			name = L["Toggle Hermes On/Off"],
			get = function() return Hermes.db.profile.enabled end,
			order = 1,
			set = "OnSetEnabled",
        },
    },
}

function Hermes:LoadDefaultOptions()
	local defaults = {
		profile = {
			enabled = true,
			enableparty = false,
			configMode = true,
			sender = { enabled = true },
			receiver = { enabled = true },
			spells = {},
			items = {},
			plugins = {},
			pluginState = { --stores which plugins are enabled or disabled
				["HermesUI"] = true --make the built in UI enabled by default
			},
		},
	}
	self.db = LIB_AceDB:New("HermesDB2", defaults)
end

function Hermes:LoadBlizOptions()
	optionTables = {
		handler = Hermes,
		type="group",
		childGroups = "tab",
		name = "Hermes",
		args = {
			General = {
				name = L["General"],
				type="group",
				order = 0,
				args = {
					General = {
						type = "group",
						name = L["General"],
						inline = true,
						order = 0,
						args = {
							EnableHermes = {
								type = "toggle",
								name = L["Enable Hermes"],
								width = "full",
								get = function(info) return self.db.profile.enabled end,
								order = 0,
								set = "OnSetEnabled",
							},
						},
					},
					Communication = {
						type = "group",
						name = L["Communication"],
						inline = true,
						order = 1,
						disabled = function() return self.db.profile.enabled == false end,
						args = {
							ConfigMode = {
								type = "toggle",
								name = L["Config Mode"],
								width = "normal",
								disabled = function() return self.db.profile.enabled == false end,
								get = function(info) return self.db.profile.configMode end,
								order = 5,
								set = "OnSetTestMode",
							},
							Spacer1A = {
								type = "description",
								name = "",
								width = "double",
								order = 10,
							},
							Spacer1B = {
								type = "description",
								name = "",
								width = "full",
								order = 15,
							},
							EnableSender = {
								type = "toggle",
								name = L["Enable Sending"],
								width = "normal",
								get = function(info) return Hermes.db.profile.sender.enabled end,
								disabled = function() return self.db.profile.configMode == true or self.db.profile.enabled == false end,
								order = 20,
								set = "OnSetEnableSender",
							},
							Spacer2A = {
								type = "description",
								name = "",
								width = "double",
								order = 25,
							},
							Spacer2B = {
								type = "description",
								name = "",
								width = "full",
								order = 30,
							},
							EnableReceiver = {
								type = "toggle",
								name = L["Enable Receiving"],
								width = "normal",
								get = function(info) return Hermes.db.profile.receiver.enabled end,
								disabled = function() return self.db.profile.configMode == true or self.db.profile.enabled == false end,
								order = 35,
								set = "OnSetEnableReceiver",
							},
							Spacer3A = {
								type = "description",
								name = "",
								width = "double",
								order = 40,
							},
							Spacer3B = {
								type = "description",
								name = "",
								width = "full",
								order = 45,
							},
							EnableParty = {
								type = "toggle",
								name = L["Enable Party Support"],
								get = function(info) return Hermes.db.profile.enableparty end,
								disabled = function() return self.db.profile.configMode == true or self.db.profile.enabled == false end,
								order = 50,
								set = "OnSetEnablePartySupport",
							},
							Spacer4A = {
								type = "description",
								name = "",
								width = "double",
								order = 55,
							},
							Spacer4B = {
								type = "description",
								name = "",
								width = "full",
								order = 60,
							},
						},
					},
					PluginList = {
						name = L["Registered Plugins"],
						type = "group",
						inline = true,
						order = 4,
						disabled = function() return self.db.profile.enabled == false end,
						args = {
						},
					},
					UpgradeOptions = {
						name = L["Upgrade Options"],
						type = "group",
						inline = true,
						order = 5,
						args = {
							RefreshButton = {
								type = "execute",
								name = L["Update Ability Defaults"],
								width = "normal",
								order = 1,
								disabled = function() return self.db.profile.enabled == false end,
								desc = "",
								func = function()
									Hermes:SetupNewProfileSpells()
									Hermes:SetupNewProfileItems()
								end,
							},
							Spacer1A = {
								type = "description",
								name = "",
								width = "double",
								order = 2,
							},
							Spacer1B = {
								type = "description",
								name = "",
								width = "full",
								order = 3,
							},
						},
					},
				},
			},
			Spells = {
				name=L["Spells"],
				type="group",
				childGroups="tab",
				order = 1,
				disabled = function() return self.db.profile.enabled == false end,
				args = { }
			},
			Items = {
				name=L["Items"],
				type="group",
				childGroups="tab",
				order = 2,
				disabled = function() return self.db.profile.enabled == false end,
				args = { }
			},
		},
	}
	
	--self:UpdateBlizOptionsTableReferences()
	LIB_AceConfig:RegisterOptionsTable(AddonName, optionTables)
	LIB_AceConfig:RegisterOptionsTable(AddonName.." ".."Command Line", slashCommands, {"hermes"})
	self.blizzOptionsFrame = LIB_AceConfigDialog:AddToBlizOptions(AddonName, AddonName)
	optionTables.args.Profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	optionTables.args.Profiles.order = -1
	LIB_AceConfigDialog:SetDefaultSize(AddonName, 700, 550)
end

function Hermes:UpdateBlizOptionsTableReferences()
	self:BlizOptionsTable_Spells()
	self:BlizOptionsTable_Items()
	Hermes:BlizOptionsTable_PluginList()
	self:BlizOptionsTable_Plugins()
end

function Hermes:BlizOptionsTable_Spells()
	if(optionTables.args.Spells.args) then
		wipe(optionTables.args.Spells.args)
	end

	--setup basic class structure
	for classEnum, className in ipairs(Hermes.CLASS_ENUM) do
		
		local classGroup = {
			type = "group",
			inline = true,
			name = Hermes:GetColorCodedClassText(Hermes:GetLocalizedClassName(className), className),
			order = classEnum,
			args = { }
		}
		
		local groupHasSpells = false
		for i, spell in ipairs(self.db.profile.spells) do
			if(spell.class == className) then
				groupHasSpells = true
				classGroup.args[tostring(spell.id)] = {
					type = "group",
					inline = true,
					name = "",
					order = i,
					args = {
						name = {
							type = "toggle",
							name = "|T"..spell.icon..":0:0:0:0|t "..spell.name,
							order = 0,
							width = "double",
							desc = tostring(spell.id),
							get = function() return spell.enabled end,
							set =
								function(info, value)
									if(spell.enabled) then
										spell.enabled = false
										if(Hermes.Receiving == true) then
											self:StopTrackingAbility(spell)
										end
									else
										spell.enabled = true
										if(Hermes.Receiving == true) then
											self:StartTrackingAbility(spell)
										end
										
									end
								end,
						},
						DeleteButton = {
							type = "execute",
							name = L["Delete"],
							width = "normal",
							order = 2,
							desc = spell.name,
							func = function()
								self:DeleteSpell(spell)
								LIB_AceConfigRegistry:NotifyChange(AddonName)
								Hermes:BlizOptionsTable_Spells()
							end,
							confirm = function() return L["Spell will be deleted. Continue?"] end,
						}
					}
				}
			end
		end
		
		if(groupHasSpells) then
			optionTables.args.Spells.args[className] = classGroup
		end
	end
	
	optionTables.args.Spells.args.Create = {
		type="group",
		name= L["Add Spell"],
		inline = true,
		order = 0,
		args = {
			Instructions = {
				type = "description",
				name = "",
				width = "full",
				order = 0,
			},
			ClassSelect = {
				type = "select",
				name = L["Class"],
				order = 1,
				style = "dropdown",
				width = "normal",
				values = function()
					local values = {}
					values[L["-- Select --"]] = L["-- Select --"]
					for i, classFileName in ipairs(Hermes.CLASS_ENUM) do
						values[classFileName] = Hermes:GetColorCodedClassText(Hermes:GetLocalizedClassName(classFileName), classFileName)
					end
					return values
				end,
				get = function(info) return newSpellTemplate.class end,
				set = function(info, value) newSpellTemplate.class = value end,
			},
			SpellNameOrId = {
				type = "input",
				name = L["Name or ID"],
				order = 2,
				width = "normal",
				get = function(info)
					if(newSpellTemplate.name ~= nil) then
						return newSpellTemplate.name
					elseif(newSpellTemplate.id ~= nil) then
						return string.format("%.0f", newSpellTemplate.id)
					else
						return ""
					end
				end,
				set = function(info, value)
					local n = tonumber(self:Trim(value))
					local s = self:Trim(value)
					if(strlen(s) == 0) then s = nil end
					if(n and n ~= 0) then
						newSpellTemplate.id = n
						newSpellTemplate.name = nil
					else
						newSpellTemplate.id = nil
						newSpellTemplate.name = s
					end
				end,
			},
			AddSpellButton = {
				type = "execute",
				name = L["Add Spell"],
				width = "normal",
				order = 3,
				func = function()
					if Hermes:AddSpell(tonumber(newSpellTemplate.id), newSpellTemplate.name, newSpellTemplate.class) then
						--reset the entered data
						newSpellTemplate.id = nil
						newSpellTemplate.name = nil
						
						--force the config window to update with latest spell info
						LIB_AceConfigRegistry:NotifyChange(AddonName)
						Hermes:BlizOptionsTable_Spells()
					end
				end,
				disabled = function()
					--ensure that a class is selected and that spell evaluates to a spell id or name
					return newSpellTemplate.class == L["-- Select --"] or (not newSpellTemplate.name and not newSpellTemplate.id)
				end,
			},
		}
	}
end

function Hermes:BlizOptionsTable_Items()
	if(optionTables.args.Items.args) then
		wipe(optionTables.args.Items.args)
	end

	--setup basic class structure
	for classEnum, className in ipairs(Hermes.CLASS_ENUM) do
		
		local classGroup = {
			type = "group",
			inline = true,
			name = Hermes:GetColorCodedClassText(Hermes:GetLocalizedClassName(className), className),
			order = classEnum,
			args = { }
		}
		
		local groupHasItems = false
		for i, item in ipairs(self.db.profile.items) do
			if(item.class == className) then
				groupHasItems = true
				classGroup.args[tostring(item.id)] = {
					type = "group",
					inline = true,
					name = "",
					order = i,
					args = {
						name = {
							type = "toggle",
							name = "|T"..item.icon..":0:0:0:0|t "..item.name,
							order = 0,
							width = "double",
							desc = tostring(Hermes:DecodeTrackerIdentifier(item.id)),
							get = function() return item.enabled end,
							set = function(info, value)
								if(item.enabled) then
									item.enabled = false
									if(Hermes.Receiving == true) then
										self:StopTrackingAbility(item)
									end
								else
									item.enabled = true
									if(Hermes.Receiving == true) then
										self:StartTrackingAbility(item)
									end
								end
							end,
						},
						DeleteButton = {
							type = "execute",
							name = L["Delete"],
							width = "normal",
							order = 2,
							desc = item.name,
							func = function()
								self:DeleteItem(item)
								LIB_AceConfigRegistry:NotifyChange(AddonName)
								Hermes:BlizOptionsTable_Items()
							end,
							confirm = function() return L["Item will be deleted. Continue?"] end,
						}
					}
				}
			end
		end
		
		if(groupHasItems) then
			optionTables.args.Items.args[className] = classGroup
		end
	end
	
	optionTables.args.Items.args.Create = {
		type ="group",
		name = L["Add Item"],
		inline = true,
		order = 0,
		args = {
			Instructions = {
				type = "description",
				name = "",
				width = "full",
				order = 0,
			},
			ClassSelect = {
				type = "select",
				name = L["Class"],
				order = 1,
				style = "dropdown",
				width = "normal",
				values = function()
					local values = {}
					values[L["-- Select --"]] = L["-- Select --"]
					for i, classFileName in ipairs(Hermes.CLASS_ENUM) do
						values[classFileName] = Hermes:GetColorCodedClassText(Hermes:GetLocalizedClassName(classFileName), classFileName)
					end
					return values
				end,
				get = function(info) return newItemTemplate.class end,
				set = function(info, value) newItemTemplate.class = value end,
			},
			ItemNameOrId = {
				type = "input",
				name = L["Name or ID"],
				order = 2,
				width = "normal",
				get = function(info)
					if(newItemTemplate.name ~= nil) then
						return newItemTemplate.name
					elseif(newItemTemplate.id ~= nil) then
						return string.format("%.0f", newItemTemplate.id)
					else
						return ""
					end
				end,
				set = function(info, value)
					local n = tonumber(self:Trim(value))
					local s = self:Trim(value)
					if(strlen(s) == 0) then s = nil end
					if(n and n ~= 0) then
						newItemTemplate.id = n
						newItemTemplate.name = nil
					else
						newItemTemplate.id = nil
						newItemTemplate.name = s
					end
				end,
			},
			AddItemButton = {
				type = "execute",
				name = L["Add Item"],
				width = "normal",
				order = 3,
				func = function()
					if Hermes:AddItem(tonumber(newItemTemplate.id), newItemTemplate.name, newItemTemplate.class) then
						--reset the entered data
						newItemTemplate.id = nil
						newItemTemplate.name = nil
						
						--force the config window to update with latest spell info
						LIB_AceConfigRegistry:NotifyChange(AddonName)
						Hermes:BlizOptionsTable_Items()
					end
				end,
				disabled = function()
					--ensure that a class is selected and that spell evaluates to a spell id or name
					return newItemTemplate.class == L["-- Select --"] or (not newItemTemplate.name and not newItemTemplate.id)
				end,
			}
		}
	}
end

function Hermes:BlizOptionsTable_PluginList()
	if optionTables.args.General.args.PluginList.args then
		wipe(optionTables.args.General.args.PluginList.args)
	end
	
	for name, _ in pairs(Hermes.Plugins) do
		optionTables.args.General.args.PluginList.args[name] = {
			type = "toggle",
			name = name,
			width = "full",
			get = function(info)
				return self.db.profile.pluginState[name] == true
			end,
			order = 5,
			set = function(info, value)
				--start or stop the plugin based on change in state
				if value == false then
					Hermes:DisablePlugin(name)
					self.db.profile.pluginState[name] = false
					--reload the options
					LIB_AceConfigRegistry:NotifyChange(AddonName)
					self:BlizOptionsTable_Plugins()
				else
					--for now, just restart hermes until I figure out a better solution to handle an addon being enabled while hermes is already running
					Hermes:Shutdown()
					self.db.profile.pluginState[name] = true
					Hermes:Startup()
				end
			end,
		}
	end
end

function Hermes:BlizOptionsTable_Plugins()
	for name, plugin in pairs(Hermes.Plugins) do
		--reset prior table if it existed
		if optionTables.args[name] then
			optionTables.args[name] = nil
		end
				
		if self.db.profile.pluginState[name] == true then
			--the selected plugin is enabled
			local pluginOptions = nil
			--see if it has an options callback
			if plugin.OnGetBlizzOptionsTable then
				pluginOptions = plugin.OnGetBlizzOptionsTable()
			end
				
			--see if the options callback returned a table
			if pluginOptions then
				--manually control how to disable it
				pluginOptions.disabled = function() return self.db.profile.enabled == false end
				optionTables.args[name] = pluginOptions
			end
		end
	end
end

function Hermes:OnSetEnabled(info, value)
	self.db.profile.enabled = value
	if(value == true) then
		Hermes:Startup()
	else
		Hermes:Shutdown()
	end
end

function Hermes:OnSetTestMode(info, value)
	self.db.profile.configMode = value
	
	if self.db.profile.configMode == true then
		Hermes:StartTestMode()
	else
		Hermes:StopTestMode()
	end
end

function Hermes:OnSetEnablePartySupport(info, value)
	self.db.profile.enableparty = value
	self:UpdateCommunicationsStatus()
end

function Hermes:OnSetEnableSender(info, value)
	self.db.profile.sender.enabled = value
	self:UpdateCommunicationsStatus()
end

function Hermes:OnSetEnableReceiver(info, value)
	self.db.profile.receiver.enabled = value
	self:UpdateCommunicationsStatus()
end
